/* colan.pl */
 
/* Suitable extensions of ADAM for handling constraints
   Creation of the appropriate metaclasses and classes
*/
 
 
%%% See section 3.3.1. of the thesis
colan_meta_class :-
	new([colan_meta_class, [
 	     is_a([keyed_meta_class]),
		 method((new(global,[],[plog,plog],[],[Class,AttributesIn]) :-
				route(Class,AttributesIn,AttributesOut),
				new([Class,AttributesOut]) => super,
				init_counter => Class,
				create_method_templates(Class,AttributesIn),
				inherit_pending_constraints(Class)
			))
		]]) => meta_class.
 
%%% See section 3.3.2. of the thesis
colan_class :-
	new([colan_class, [
		is_a([keyed_class_mixin]),
		slot(slot_tuple(is_a, system, set, optional, object)),
		slot(slot_tuple(slot_desc, system, set, optional, slot_desc_tuple)),
		slot(slot_tuple(key_desc, global, single, optional, plog)),
		slot(slot_tuple(constraints, global, set, optional, plog)),
		slot(slot_tuple(new_constraints, global, set, optional, plog)),
		slot(slot_tuple(delete_constraints, global, set, optional, plog)),
		slot(slot_tuple(put_constraints, global, set, optional, plog)),
		slot(slot_tuple(update_constraints, global, set, optional, plog)),
		slot(slot_tuple(delete_slot_constraints, global, set, optional, plog)),
		slot(slot_tuple(counter, global, single, total, integer)),
		method((inc_counter(global,[],[],[],[]) :-
				message_recipient(Class),
				get_all_super_classes(Class,Supers),
				findall(Super,
					(member(Super,[Class|Supers]),
					 get_counter(Old) => Super,
					 New is Old + 1,
					 update_counter([Old,New]) => Super),
					_)
			)),
		method((dec_counter(global,[],[],[],[]) :-
				message_recipient(Class),
				get_all_super_classes(Class,Supers),
				findall(Super,
					(member(Super,[Class|Supers]),
					 get_counter(Old) => Super,
					 New is Old - 1,
					 update_counter([Old,New]) => Super),
					_)
			)),
		method((init_counter(global,[],[],[],[]) :-
				message_recipient(Class),
				put_counter([0]) => Class
			)),
		method((new(global,[],[plog,plog],[],[Id,Attributes]) :-
				message_recipient(Class),
				(check_constraints(new,Class,Id,Attributes) ->
					(new([Id,[deletion_status([no])|Attributes]]) => super,
				 	 inc_counter => Class))
	     		)),
		method((put_method(global,[],[plog],[],[(Head :- Body)]) :-
				message_recipient(Class),
				(member_code(check_constraints(_,_,_,_,_),Body) ->
					put_method([(Head :- Body)]) => super;
					(functor(Head,Method_name,_),
					 (not(is_it_defined(Method_name,Class)) ->
					 	(get_template(Head,Template),
					 	 HeadCall =.. [Method_name,Template],
					 	 NewBody = ((HeadCall => super),Body),
					 	 put_method([(Head :- NewBody)]) => super);
						(get_method([Method_name],(Head :- OldBody)) => Class ->
						  	((OldBody = check_constraints(_,_,_,_,_) ->
								append_code(OldBody,Body,NewBody);
								NewBody = Body),
							 (is_it_method_defined(Method_name,Class) ->
								replace_method([(Head :- NewBody)]) => super;
								put_method([(Head :- NewBody)]) => super));
						 	put_method([(Head :- Body)]) => super))))
			)),
		method((replace_method(global,[],[plog],[],[(Head :- Body)]) :-
				message_recipient(Class),
				(member_code(check_constraints(_,_,_,_,_),Body) ->
					replace_method([(Head :- Body)]) => super;
					(functor(Head,Method_name,_),
					 get_method([Method_name],(Head :- OldBody)) => Class,
					 (member_code(check_constraints(A,B,C,D,E),OldBody) ->
						NewBody = (check_constraints(A,B,C,D,E),Body);
						(get_template(Head,Template),
						 HeadCall =.. [Method_name,Template],
					 	 NewBody = ((HeadCall => super),Body))),
					 replace_method([(Head :- NewBody)]) => super))
			)),
		method((get_constraint(global,[],[],string,[Constraint]) :-
				message_recipient(Class),
				get_constraints(ConstId) => Class,
				get_constraint(Constraint) => ConstId
	     		)),
		method((put_constraint(global,[],[plog],[],[Constraint]) :-
				message_recipient(Class),
				trim(Constraint,TrimString),
				put_constraints([TrimString]) => Class
	     		)),
		method((delete_constraint(global,[],[plog],[],[Constraint]) :-
				message_recipient(Class),
				trim(Constraint,TrimString),
				get_by_constraint([TrimString],ConstId) => constraints ->
					(get_constraints(ConstId) => Class ->
						(findall(C,get_classes(C)=>ConstId,Classes),
					 	 delete_all_codes(ConstId,Classes),
					 	 delete_constraints([ConstId]) => Class,
					 	 delete => ConstId);
						(error_message(constraint_existence,
								constraint(Class,ConstId,TrimString)),
					 	 fail));
					(error_message(constraint_existence,constraint(Class,TrimString)),
					 fail)
			)),
		method((update_constraint(global,[],[plog,plog],[],[Old,New]) :-
				message_recipient(Class),
				delete_constraint([Old]) => Class,
				put_constraint(New) => Class
	     		)),
		method((get_method(global,single,[plog],plog,
					[[Method_name],(HeadOut :- Body)]) :-
				message_recipient(Class),
		   		get_method_desc(method_desc_tuple(Super,Method_name,Vis,Card,
								In,Out)) => Class,
				Head =.. [Super,Method_name,Vis,Card,In,Out,_],
				HeadOut =.. [Method_name,Vis,Card,In,Out,_],
				clause(Head,Body)
			)),
		method((display_methods(global,[],[],[],[]) :-
				message_recipient(Class),
				listing(Class)
			))
	 ]]) => colan_meta_class.
 
%%% See section 3.3.3. of the thesis
colan_object :-
	new([colan_object, [
		slot(slot_tuple(deletion_status, global, single, total, string)),
	     	method((delete(global,[],[],[],[]) :-
			  message_recipient(Id),
			  get_instance_of(Class) => Id,
			  (not(checked_for_deletion(Id)) ->
			  	(update_deletion_status([no,yes]) => Id,
				 (check_constraints(delete,Class,Id) ->
					(delete => super,
					 dec_counter => Class)));
					delete => super)
	     		))
	 ]]) => meta_class.
 
%%% See section 3.4. of the thesis
constraint_class :-
	new([constraints, [
		slot(slot_tuple(constraint, global, single, optional, plog)),
		slot(slot_tuple(classes, global, set, total, colan_class))
	]]) => class.
 
%%% Changing the default method definition for 'put_constraints'
%%% Constraints must first be translated and stored
put_constraints :-
	replace_method([(put_constraints(global,[],[plog],[],[Constraint]) :-
				message_recipient(Class),
				create_method(Class,Constraint,ConstId),
				constraints <-- ConstId
	     	)]) => colan_class.
 
 
:- colan_meta_class, colan_class, colan_object, 
   constraint_class, put_constraints.
