/* inverter.pl */

/* This file does the inversion for the symmetrical constraints,
   and also creates the slot methods (put_,update_) for the multiple
   slot constraints (constraints involving more than one slots 
*/


%%% invert_constraint(+Use,+Methods,+Constraint,-ListOfCodes,-ListOfClasses)
%%% returns the classes that need to store code for a constraint
%%% and the codes for the various methods affected
invert_constraint(Use,Methods,
		  constraint([counter(forall(Var),Var,Class)|Rest],Preds),
		  RestCodes,Classes) :- 
	findall(C,(member(counter(_,_,C),Rest),atom(C)),AllClasses),
	unique(AllClasses,RestClasses),
	not_inheritable_classes(Class,RestClasses,Classes),
	invert_constraint_aux(Use,Methods,Classes,counter(forall(Var),Var,Class),
					Rest,[],Preds,RestCodes), !.
invert_constraint(_,_,_,[],[]).

%%% Auxiliayr predicate for inverting a constraint
invert_constraint_aux(_,_,_,_,[],_,_,[]).
invert_constraint_aux(Use,Methods,Classes,counter(forall(Var),Var,Class),
			[counter(Pred,Var1,Class1)|Tail1],
			Temp,Preds,[compiled(Class1,NewCode)|TailCode]) :- 
	member(Class1,Classes),
	reverse(Temp,RTemp),
	Pred =.. [Operation,Var1,Value],
	NewPred =.. [Operation,Var,Value],
	append([counter(forall(Var1),Var1,Class1)|RTemp],
	       [counter(NewPred,Var,Class)|Tail1],
	       NewVars),
	findall(OneNewPred,(member(OnePred,Preds),
			    invert_predicates(OnePred,Var,Var1,OneNewPred)),
		NewPreds), !,
	generate_all(Use,Methods,Class1,constraint(NewVars,NewPreds),NewCode),
	invert_constraint_aux(Use,Methods,Classes,counter(forall(Var),Var,Class),Tail1,
				[counter(Pred,Var1,Class1)|Temp],Preds,TailCode).
invert_constraint_aux(Use,Methods,Classes,First,[Head1|Tail1],Temp,Preds,RestCodes) :- 
	invert_constraint_aux(Use,Methods,Classes,First,Tail1,[Head1|Temp],Preds,RestCodes).

%%% invert_predicates(+PredicateList,+Var1,+Var2,-InversedPredicatesList)
%%% returns a list with all the predicates of the constraint inversed
invert_predicates([],_,_,[]).
invert_predicates([Pred|Tail],Var1,Var2,[NewPred|NewTail]) :- 
	Pred =.. [Operation,Op1,Op2],
	invert_operands(Op1,Op2,Var1,Var2,NewOp1,NewOp2),
	invert_operation(Operation,InvertedOperation),
	NewPred =.. [InvertedOperation,NewOp1,NewOp2],
	invert_predicates(Tail,Var1,Var2,NewTail).

%%% inverts the operands of an operation
invert_operands(slot(Slot1,Var1),slot(Slot2,Var2),Var1,Var2,
		slot(Slot2,Var2),slot(Slot1,Var1)) :- !.
invert_operands(instance(Var1),instance(Var2),Var1,Var2,
		instance(Var2),instance(Var1)).

%%% inverses the operation
invert_operation(>,=<) :- !.
invert_operation(<,>=) :- !.
invert_operation(>=,<) :- !.
invert_operation(=<,>) :- !.
invert_operation(in,not_in) :- !.
invert_operation(not_in,in) :- !.
invert_operation(A,A).
