/* optimizer.pl */

/* This file optimizes the code generated from CoLan */


%%% takes a list of codes and returns them optimized
optimizer([],[]).
optimizer([H1|T1],[H2|T2]) :- 
	optimize(H1,H2),
	optimizer(T1,T2).

%%% optimizes one piece of code
optimize((H,T),Opt) :- !,
	optimize(T,(H),ROpt,and),
	reverse_code(ROpt,Opt,and).
optimize((H;T),Opt) :- 
	not(((H = (_->_));(H = (_>>>_)))), !,
	optimize(T,(H),ROpt,or),
	reverse_code(ROpt,Opt,or).
optimize((Head :- Body),(Head :- Opt)) :- !,
	optimize(Body,Opt).
optimize(H,H).

%%% Auxiliary predicate for optimize
optimize((H,T),Temp,Result,and) :- 
	member_if_code(H,Temp,Then,and), !,
	optimize(T,Then,Result,and).
optimize((H,T),Temp,Result,and) :- !,
	optimize(T,(H,Temp),Result,and).
optimize((H;T),Temp,Result,or) :- 
	not(((H = (_->_));(H = (_>>>_)))),
	member_if_code(H,Temp,Then,or), !,
	optimize(T,Then,Result,or).
optimize((H;T),Temp,Result,or) :- 
	not(((H = (_->_));(H = (_>>>_)))), !,
	optimize(T,(H;Temp),Result,or).
optimize(H,Temp,Then,Con) :- 
	member_if_code(H,Temp,Then,Con), !.
optimize(H,Temp,(H,Temp),and) :- !.
optimize(H,Temp,(H;Temp),or).

%%% Takes two pieces of code and returns then concatenated
%%% in optimized form
member_if_code((H :- B1),((H :- B2)),((H :- B)),and) :- !,
	optimize((B1,B2),B).
member_if_code((H :- B1),((H :- B2)),((H :- B)),or) :- !,
	optimize((B1;B2),B).
member_if_code((H->T1;Else),((H->T2;Else)),((H->T;Else)),Con) :- !,
	append_code(T2,T1,T,Con).
member_if_code((H>>>T1),((H>>>T2)),((H>>>T)),Con) :- !,
	append_code(T2,T1,T,Con).
member_if_code((H1 -> T;E),((H2 -> T;E)),((H -> T;E)),Con) :- !,
	append_code(H1,H2,H,Con).
member_if_code((H1>>>T),((H2>>>T)),((H>>>T)),Con) :- !,
	append_code(H1,H2,H,Con).
member_if_code((H :- B1),((H :- B2),Tail),((H :- B),Tail),and) :- !,
	optimize((B1,B2),B).
member_if_code((H :- B1),((H :- B2),Tail),((H :- B);Tail),or) :- !,
	optimize((B1;B2),B).
member_if_code((H->T1;Else),((H->T2;Else),Tail),((H->T;Else),Tail),and) :- !,
	append_code(T2,T1,T,and).
member_if_code((H>>>T1),((H>>>T2),Tail),((H>>>T),Tail),and) :- !,
	append_code(T2,T1,T,and).
member_if_code((H->T1;Else),((H->T2;Else);Tail),((H->T;Else);Tail),or) :- !,
	append_code(T2,T1,T,or).
member_if_code((H>>>T1),((H>>>T2);Tail),((H>>>T);Tail),or) :- !,
	append_code(T2,T1,T,or).
member_if_code((H1 -> T;E),((H2 -> T;E),Tail),((H -> T;E),Tail),and) :- !,
	append_code(H1,H2,H,and).
member_if_code((H1>>>T),((H2>>>T),Tail),((H>>>T),Tail),and) :- !,
	append_code(H1,H2,H,and).
member_if_code((H1 -> T;E),((H2 -> T;E);Tail),((H -> T;E);Tail),or) :- !,
	append_code(H1,H2,H,or).
member_if_code((H1>>>T),((H2>>>T);Tail),((H>>>T);Tail),or) :- !,
	append_code(H1,H2,H,or).
member_if_code(H,(Head,Tail),(Head,Then),and) :- 
	member_if_code(H,Tail,Then,and).
member_if_code(H,(Head;Tail),(Head;Then),or) :- 
	member_if_code(H,Tail,Then,or).

%%% It inputs a parse tree and returns an optimized one
optimize_tree([],[]) :- !.
optimize_tree(constraint([counter(Var<1,Var,Class)],Preds),
	      constraint([counter(forall(Var),Var,Class)],NoPreds)) :- !,
	negate(Preds,NoPreds).
optimize_tree(constraint([counter(forall(Var1),Var1,Class1),
			  counter(forall(Var2),Var2,Class2)],Preds),
	      constraint([counter(forall(Var1),Var1,Class1),
			  counter(Var2<1,Var2,Class2)],NoPreds)) :- !,
	negate(Preds,NoPreds).
optimize_tree(constraint([counter(forall(Var1),Var1,Class)],Preds),
	      constraint([counter(forall(Var1),Var1,InnerClass),
			  counter(Var2<1,Var2,Class)],NoNewPreds)) :- 
	not(atom(Class)), !,
	find_inner_class(Class,InnerClass),
	create_variable(Var1,Var2),
	replace_variable(Var1,Var2,Preds,NewPreds),
	negate(NewPreds,NoNewPreds).
optimize_tree(constraint([counter(forall(Var1),Var1,Class)],[[Pred]]),
	      constraint([counter(forall(Var1),Var1,Class),
			  counter(Var2<1,Var2,ComplexClass)],[[NoNewPred]])) :- 
	create_variable(Var1,Var2),
	Pred =.. [Operator,Op1,Op2],
	select_complex(Op1,Op2,Var2,ComplexClass,NewOp1,NewOp2), !,
	negation(Operator,NotOperator),
	NoNewPred =.. [NotOperator,NewOp1,NewOp2].
optimize_tree(A,A).

%%% It negates a list of predicates
negate(Preds,NoPreds) :- 
	findall(Item,comb(Preds,Item),NoPreds).

%%% Auxiliary predicate for negate
comb([H|T],[NoOne|T1]) :- 
	member(One,H),
	negate_one(One,NoOne),
	comb(T,T1).
comb([],[]).

%%% It negates one predicate
negate_one(Pred,NoPred) :- 
	Pred =.. [Operation,Op1,Op2],
	negation(Operation,NotOperation),
	NoPred =.. [NotOperation,Op1,Op2].


%%% select_complex(+Op1,+Op2,+Var2,-ComplexClass,-NewOp1,-NewOp2)
%%% finds which if the Op1 and Op2 is a complex function and strips all the
%%% functions but the first.  It returns the stripped functions as ComplexClass
%%% and sticks another variable to the previously complex function
select_complex(slot(Slot1,Var1),slot(Slot2,ComplexClass),Var2,ComplexClass,
			slot(Slot1,Var1),slot(Slot2,Var2)) :- 
	atom(Var1), not(atom(ComplexClass)), !.
select_complex(slot(Slot2,ComplexClass),slot(Slot1,Var1),Var2,ComplexClass,
			slot(Slot2,Var2),slot(Slot1,Var1)) :- 
	atom(Var1), not(atom(ComplexClass)), !.
