/* parser.pl */

/* This is the parser for the CoLan language */


constraint(Class,constraint(Counters,[])) -->
	quantification(Counters), 
	{!, check_first_counter(Class,Counters)}.
constraint(Class,constraint(Counters,Preds)) -->
	quantification(Counters),
	[such,that],
	predicates(Preds), 
	{!, check_first_counter(Class,Counters)}.
constraint(Class,constraint(Counters,Preds)) -->
	quantification(Counters),
	predicates(Preds),
	{check_first_counter(Class,Counters)}.

quantification([counter(Range,Counter,Class)|Counters]) -->
	quantifier(Range,Counter,Class),
	quantification(Counters), {!}.
quantification([counter(Range,Counter,Class)]) -->
	quantifier(Range,Counter,Class).

predicates([Preds]) -->
	conjuction(Preds), {!}.
predicates([Pred|Preds]) -->
	conjuction(Pred), 
	[or], 
	predicates(Preds).

conjuction([Pred|Preds]) -->
	predicate(Pred), 
	[and], 
	conjuction(Preds), {!}.
conjuction([Pred]) -->
	predicate(Pred).

quantifier(forall(Instances),Instances,CompClassOut) -->
	[forall,Instances,in],
	compound_class(CompClassIn),
	{check_class(CompClassIn),
	 replace_all_variables(CompClassIn,CompClassOut),
	 bond(Instances,CompClassIn), !}.
quantifier(<(Instance,1),Instance,CompClassOut) -->
	[not,exists,Instance,in],
	compound_class(CompClassIn),
	{check_class(CompClassIn),
	 replace_all_variables(CompClassIn,CompClassOut),
	 bond(Instance,CompClassIn), !}.
quantifier(>=(Instance,1),Instance,CompClassOut) -->
	[exists,Instance,in],
	compound_class(CompClassIn),
	{check_class(CompClassIn),
	 replace_all_variables(CompClassIn,CompClassOut),
	 bond(Instance,CompClassIn), !}.
quantifier(Range,Instances,CompClassOut) -->
	[exist],
	num_quantifier(Range),
	[Instances,in],
	compound_class(CompClassIn),
	{check_class(CompClassIn),
	 replace_all_variables(CompClassIn,CompClassOut),
	 unify_counter(Instances,Range),
	 bond(Instances,CompClassIn), !}.

num_quantifier(Range) -->
	numerical_quantifier(Range).

numerical_quantifier(=(_,N)) -->
	[only,N],
	{integer(N), !}.
numerical_quantifier(=(_,N)) -->
	[exactly,N],
	{integer(N), !}.
numerical_quantifier(=(_,N)) -->
	[=,N],
	{integer(N)}.
numerical_quantifier(>(_,N)) -->
	[more,than,N],
	{integer(N), !}.
numerical_quantifier(>(_,N)) -->
	[>,N],
	{integer(N)}.
numerical_quantifier(>=(_,N)) -->
	[no,less,than,N],
	{integer(N), !}.
numerical_quantifier(>=(_,N)) -->
	[at,least,N],
	{integer(N), !}.
numerical_quantifier(>=(_,N)) -->
	[>=,N],
	{integer(N)}.
numerical_quantifier(<(_,N)) -->
	[less,than,N],
	{integer(N), !}.
numerical_quantifier(<(_,N)) -->
	[<,N],
	{integer(N)}.
numerical_quantifier(=<(_,N)) -->
	[no,more,than,N],
	{integer(N), !}.
numerical_quantifier(=<(_,N)) -->
	[at,most,N],
	{integer(N), !}.
numerical_quantifier(=<(_,N)) -->
	[=<,N],
	{integer(N)}.

compound_class(slot(Slot,Arg)) -->
	[Slot,of],
	compound_class(Arg), {!}.
compound_class(Class) -->
	[Class],
	{atom(Class)}.

predicate(Pred) -->
	multi_operand(Op1),
	rel_con(Connector),
	multi_operand(Op2),
	{Pred =.. [Connector,Op1,Op2]}.

multi_operand(Op) -->
	operand(Op,_).
multi_operand(Op) -->
	c_operand(Op).

operand(slot(Slot,Instance),Class) -->
	[Slot,of,Instance],
	{check_instance(Instance,Class),check_slot(Slot,Class), !}.
operand(slot(Slot,Arg),Class) -->
	[Slot,of], operand(Arg,PrevClass),
	{check_slot_of_arg(Arg,PrevClass,Class),check_slot(Slot,Class), !}.
operand(instance(Instance),Class) -->
	[Instance],
	{check_instance(Instance,Class), !}.
operand(N,_) -->
	[N], {constant(N)}.

c_operand(Op) -->
	operand(Op1,_),
	operation(Operation),
	operand(Op2,_),
	{Op =.. [Operation,Op1,Op2]}.


operation(S) -->
	[S], {oper_symbol(S)}.

oper_symbol(+).
oper_symbol(-).
oper_symbol(*).
oper_symbol(/).

rel_con(COut) -->
	[CIn], {connector(CIn,COut), !}.
rel_con(in) -->
	[is,in], {!}.
rel_con(not_in) -->
	[not,in], {!}.

connector(=,equal).
connector(<>,not_equal).
connector(C,C) :-
	connector(C).

connector(>).
connector(<).
connector(>=).
connector(=<).

