/* parser_aux.pl */

/* This file consists of auxiliary predicates
   for the parser of CoLan
*/



:- dynamic bonded/2, unchecked_class/1.


%%% This predicate checks if a class exists.
%%% If not it asserts a flag that indicates that the class
%%% should be later checked.
check_class(C) :- 
	check_comp_class(C,_), !.	
check_class(C) :- 
	asserta(unchecked_class(C)).

check_comp_class(Class,Class) :- 
	atom(Class),
	get(Class) => colan_class, !.
check_comp_class(Instance,Class) :- 
	atom(Instance), !,
	bonded(Instance,Class).
check_comp_class(slot(Slot,Arg),Class) :- 
	check_comp_class(Arg,Previous),
	get_all_super_classes(Previous,Supers),
	bagof(Super,(member(Super,[Previous|Supers]),
		     get_slot_desc(slot_desc_tuple(Slot,_,_,_,_,Class)) => Super),
		_).

%%% check_slot(+Slot,+Class)
%%% checks if Slot is attached to Class or superclasses of Class
check_slot(S,C) :- 
	not(unchecked_class(C)),
	check_slot_aux(S,C), !.
check_slot(S,C) :-
	not(unchecked_class(C)), !,
	write(S),
	write(' is not a slot of class '),
	write(C),
	write('.'), nl.
check_slot(_,_).

check_slot_aux(Slot,Expr) :- 
	find_outer_class(Expr,Class),
	get_all_super_classes(Class,Supers),
	findall(Super,(member(Super,[Class|Supers]),
		       get_slot_desc(slot_desc_tuple(Slot,_,_,_,_,_)) => Super),
	        List), 
	List \== [], !.

%%% check_instance(+Var,-Class)
%%% checks if Var has been declared and it returns the class
%%% that is bound with it
check_instance(I,C) :-
	bonded(I,C).

%%% check_slot_of_arg(ComplexSlot,ComplexClass,Class)
%%% checks if the complex slot ComplexSlot is bound to the class Class
%%% the middle argument is auxilliary
check_slot_of_arg(slot(Slot,_),ComplexPrevious,Class) :- 
	find_outer_class(ComplexPrevious,Previous),
	get_all_super_classes(Previous,Supers),
	bagof(Super,(member(Super,[Previous|Supers]),
		     get_slot_desc(slot_desc_tuple(Slot,_,_,_,_,Class)) => Super),
		_).

%%% Binds the variable I to the class C.
bond(I,C) :- 
	asserta(bonded(I,C)).

%%% Just strips the variable atom from the quantification predicate
unify_counter(Var,Term) :- 
	Term =.. [_,Var,_].

%%% checks if C is either an integer or an atom
constant(C) :- 
	integer(C).
constant(C) :- 
	atom(C).

%%% checks whether the first variable quantifies over the host
%%% class or one of its set slots
check_first_counter(Class,[counter(_,_,Expr)|_]) :- 
	find_inner_class(Expr,Class), !.
check_first_counter(Class1,[counter(_,_,Class2)|_]) :- 
	write('The first variable of the constraint definition'), nl,
	write('should be of class "'),
	write(Class1),
	write('", and not of class "'),
	write(Class2),
	write('".'), nl.

