/* schema1.pl */

/* Example schema for university
   with sample constraints in it
*/


u3 :- new([person, [
         slot(slot_tuple(cname, global, single, total, string)),
	     slot(slot_tuple(sname, global, single, total, string)),
	     slot(slot_tuple(age, global, single, optional, integer)),
	     slot(slot_tuple(loves, global, set, optional, person)),
	     constraints([
			'forall p in person age of p > 0 and age of p < 120',
			'exist exactly 3 p in person such that age of p > 100'
			]),
	     key([sname, cname])
	 ]]) => colan_class.

u4 :- new([staff, [
		is_a([person]),
		slot(slot_tuple(research_interest, global, single, optional, string)),
		slot(slot_tuple(phone, global, single, optional, integer)),
		constraints([
				'exist at most 50 s in staff',
				'forall s in staff age of s < 65'
			])
	 ]]) => colan_class.

u5 :- new([ra, [
         is_a([staff]),
	     slot(slot_tuple(projects, global, set, optional, string)),
	     slot(slot_tuple(tutor, global, single, optional, staff)),
	     constraints(['forall r in ra exist exactly 2 r1 in ra such that phone of r = phone of r1'
			])
	 ]]) => colan_class.

u6 :- new([lecturer, [
         is_a([staff]),
	     slot(slot_tuple(supervisor_of, global, set, optional, ra)),
	     slot(slot_tuple(projects, global, set, optional, string)),
	     constraints(['forall l in lecturer projects of l = projects of supervisor_of of l',
			'forall l in lecturer not exists r in supervisor_of of l such that research_interest of l <> research_interest of tutor of r',
			'forall l in lecturer not exists s in supervisor_of of l such that age of l =< age of s',
			'forall l in lecturer not exists s in staff such that phone of l = phone of s'
			])
	 ]]) => colan_class.

u7 :- new([student, [
         is_a([person]),
	     slot(slot_tuple(year, global, single, total, integer)),
	     constraints(['forall e in student age of e + year of e >= 19',
			  'forall s in student year of s > 0 and year of s < 5',
			  'forall s in student year of s < age of s'
			])
	 ]]) => colan_class.

u8 :- new([undergrad, [
             is_a([student])
	 ]]) => colan_class.

u9 :- new([postgrad, [
         is_a([student]),
	     slot(slot_tuple(subject, global, single, total, string)),
	     slot(slot_tuple(supervisor, global, single, optional, lecturer)),
	     constraints(['exist at most 10 p in postgrad'])
	 ]]) => colan_class.

u10 :- new([pgra, [
         is_a([postgrad, ra]),
	     slot(slot_tuple(percentra, global, single, optional, integer)),
	     constraints(['forall p in pgra exist at least 2 p1 in pgra such that phone of p = phone of p1',
			  'forall p in pgra year of p < 4'
			])
	 ]]) => colan_class.

%% Declare the schema as an in-memory database.
declare_schema :-
	u3, u4, u5, u6,
	u7, u8, u9, u10.

:- declare_schema.
