/* storer.pl */

/* 
   This file has as input the compiled constraints
   and it stores them in the schema using the right methods
   and/or slots
*/


%%% store_codes(+Current_class,+ListOfCodes,+ConstId)
%%% stores the codes of the list into the appropriate classes
%%% and slots along with the constraint identifier
store_codes(_,[],_) :- !.
store_codes(Current,[compiled(Class,Codes)|Rest],ConstId) :-
	store_codes(Current,Class,Codes,ConstId),
	store_codes(Current,Rest,ConstId).

store_codes(_,_,[],_).
store_codes(Current,Class,[method(Method,Code)|Rest],ConstId) :- 
	store_code(Current,Class,Method,Code,ConstId),
	store_codes(Current,Class,Rest,ConstId).

%%% store_code(+CurrentClass,+Class,+Method,+Code,+ConstId)
%%% stores the Code in the slot corresponding to Code, along with the constraint
%%% identifier ConstId, in Class.
%%% If Class does not exist the code is stored in memory in order to be stored later
%%% when the Class is created.
store_code(_,Class,Method,Code,ConstId) :-
	exists_class(Class), !,
	message_name(Method,Name),
	concat(Name,'_constraints',Msg),
	create_message_head(put_,Msg,[constraint(ConstId,Code)],Call),
	call((Call => Class)).
store_code(Current,Class,Method,Code,ConstId) :- !,
	asserta(constraint_inheritance_pending(Class,Current,Method,constraint(ConstId,Code))).

%%% creates the method templates for all the methods of a class,
%%% i.e. adds the appropriate check_constraint calls
create_method_templates(Class,Attributes) :- 
	findall(Method,(member(slot(slot_tuple(Slot,_,_,_,_)),Attributes),
		   member(Method,[put_,update_,delete_]),
		   create_method_template(Method,Slot,Class)),
		_).

%%% auxiliary predicate for create_method_templates
create_method_template(Method,Slot,Class) :-
	create_method_head(Method,Class,Slot,Template,Head),
	message_name(Method,Name),
	concat(Name,'_constraints',Msg),
	NewCode = check_constraints(Method,Slot,Msg,Template,Head),
	replace_method([(Head :- NewCode)]) => Class.
