/* transform.pl */

/* Transforms an atom into a list of words/atoms 
   and vice-versa 
*/


%%% transform(+String,-List)
%%% returns a list with atoms taken from the String
transform(Atom,Words) :- 
	name(Atom,CList),
	chars_to_list(CList,WList),
	remove_nulls(WList,Words), !.

%%% chars_to_list(+List1,-List2)
%%% takes the List1 as a list of ASCII chars and returns
%%% an ASCII list with the words
chars_to_list([],[]).
chars_to_list([59],[]).		%% semi-colon
chars_to_list([46],[]).		%% full-stop
chars_to_list(Chars,[Word|Rest]) :- 
	chars_to_word(Chars,Word,Left),
	chars_to_list(Left,Rest).

%%% Auxilliary predicate for transformation
chars_to_word(Chars,Word,Left) :- 
	chars_to_word(Chars,[],Word,Left).

chars_to_word([],SoFar,Word,[]) :- !,
	reverse(SoFar,RSoFar),
	name(Word,RSoFar).
chars_to_word([Number1,Number2|Rest],SoFar,Word,Rest) :- 
	end_of_word(Number1), 
	end_of_word(Number2), !,
	chars_to_word([Number2|Rest],SoFar,Word,Rest). 
chars_to_word([Number|Rest],SoFar,Word,Rest) :- 
	end_of_word(Number), !,
	reverse(SoFar,RSoFar),
	name(Word,RSoFar).
chars_to_word([Number|Rest],SoFar,Word,Left) :- 
	chars_to_word(Rest,[Number|SoFar],Word,Left).

%%% word delimiters
end_of_word(32).	%% space
end_of_word(44).	%% comma
end_of_word(10).	%% new-line
end_of_word(9).		%% tab

%%% removes null atoms from the list
remove_nulls([],[]).
remove_nulls([''|T],T1) :- !,
	remove_nulls(T,T1).
remove_nulls([H|T],[H|T1]) :- 
	remove_nulls(T,T1).

%%% re_transform(ListOfAtoms,String)
%%% concatenates a list of atoms into a string, adding spaces
%%% between words
re_transform(ListOfAtoms,TrimString) :- 
	re_transform_aux(ListOfAtoms,List),
	name(TrimString,List).
	
re_transform_aux([Last],List) :- 
	name(Last,List), !.
re_transform_aux([H|T],List) :- 
	name(H,HName),
	append(HName,[32],First),
	re_transform_aux(T,Rest),
	append(First,Rest,List).

%%% removes all unnecessary spaces and other delimiters
%%% of StringIn returning StringOut
trim(StringIn,StringOut) :- 
	transform(StringIn,List),
	re_transform(List,StringOut).
